﻿using System;
using System.Data;
using System.Net;
using System.Net.Sockets;
using System.Collections;
using System.IO;
using System.Threading;
using System.Text;
using MainLibrary;
using WebLibrary;
using net.jbcg.Compression;
using System.Windows.Forms;
using System.Diagnostics;
using System.Drawing;

namespace SpeedTestDll
{
    /// <summary>
    /// The base code for running a multi-threaded async SpeedTest service for uploading and downloading data.
    /// </summary>
    public class SpeedTest
    {
        #region default_variables

        /// <summary>
        /// The number of minutes to wait before considering a connection old.
        /// </summary>
        public const int DEFAULT_OLD_CONN_MIN = 5;

        /// <summary>
        /// The number of log writes to wait before checking the log size for recycling.
        /// </summary>
        public const int CHECK_LOG_SIZE_TICKS = 25;

        /// <summary>
        /// The max log size to allow before recycling the log file.
        /// </summary>
        public const int CHECK_LOG_SIZE = 5000000;

        /// <summary>
        /// The default number of sockets that will be opened up by the server.
        /// </summary>
        public const int DEFAULT_MAX_SOCKETS = 10;

        /// <summary>
        /// The default debug setting.
        /// </summary>
        public const bool DEFAULT_DEBUG_ON = false;

        public const bool DEFAULT_PUBLIC_ON = true;

        public const float MiB_FLT = 1048576.0f;                                          //1 MiB
        public const double MiB_DBL = 1048576.0;                                          //1 MiB
        public const int MiB_INT = 1048576;                                               //1 MiB

        public const float KiB_FLT = 1024.0f;                                             //1 KiB
        public const double KiB_DBL = 1024.0;                                             //1 KiB
        public const int KiB_INT = 1024;                                                  //1 KiB

        public const float MB_FLT = 1000000.0f;                                          //1 MB
        public const double MB_DBL = 1000000.0;                                          //1 MB
        public const int MB_INT = 1000000;                                               //1 MB

        public const float KB_FLT = 1000.0f;                                             //1 KB
        public const double KB_DBL = 1000.0;                                             //1 KB
        public const int KB_INT = 1000;                                                  //1 KB

        public const int BITS_PER_BYTE = 8;                                              //1 byte, 8 bits

        public const int DEFAULT_SLOW_NETWORK_TIMEOUT_S = 10;                            //s
        public const string DEFAULT_SLOW_NETWORK_TIMEOUT_S_STR = "10";                   //s
        public const int DEFAULT_SLOW_NETWORK_TIMEOUT_MS = 10000;                        //ms
        public const string DEFAULT_SLOW_NETWORK_TIMEOUT_MS_STR = "10000";               //ms

        /// <summary>
        /// Default port number to start listening on.
        /// </summary>
        public const int DEFAULT_PORT = 49986;
        public const string DEFAULT_PORT_STR = "49986";
        public const int DEFAULT_TIMEOUT = 15000;
        public const string DEFAULT_TIMEOUT_STR = "15000";
        public const int DEFAULT_CONNECTION_TIMEOUT = 15000;
        public const string DEFAULT_CONNECTION_TIMEOUT_STR = "15000";

        public const int DEFAULT_IN_BUFFER = (MiB_INT * 2);                               //2 MiB
        public static string DEFAULT_IN_BUFFER_STR = ((MiB_INT * 2) + "");                 //2 MiB
        public const int DEFAULT_OUT_BUFFER = (MiB_INT * 2);                              //2 MiB
        public static string DEFAULT_OUT_BUFFER_STR = ((MiB_INT * 2) + "");                //2 MiB
        public const int DEFAULT_MAX_XFER = (MiB_INT * 100);                              //100 MiB
        public static string DEFAULT_MAX_XFER_STR = ((MiB_INT * 100) + "");                //100 MiB

        private const int MIN_NUM_THREADS = 5;
        private const int MAX_NUM_THREADS = 1024;
        private const int MIN_KILL_MIN = 3;
        private const int MAX_KILL_MIN = 30;
        private const int MIN_SOCKETS = 5;
        private const int MAX_SOCKETS = 1024;
        private const int MAX_IN_BUFFER_LENGTH = (MiB_INT * 5);                                //5 MiB
        private const int MIN_IN_BUFFER_LENGTH = (MiB_INT * 1);                                //1 MiB
        private const int MAX_OUT_BUFFER_LENGTH = (MiB_INT * 5);                               //5 MiB
        private const int MIN_OUT_BUFFER_LENGTH = (MiB_INT * 1);                               //1 MiB
        private const int MIN_TIMEOUT = 0;
        private const int MAX_TIMEOUT = (short.MaxValue * 2);
        private const int MIN_XFER_LIMIT = (MiB_INT * 10);                                       //10 MiB
        private const int MAX_XFER_LIMIT = (MiB_INT * 100);                                      //100 MiB
        private const int MIN_PORT = 0;
        private const int MAX_PORT = (short.MaxValue * 2);

        public const string DEFAULT_SPEED_TEST_REGISTRATION_URL = "http://middlemind.biz:83/vbsvc/Service1.asmx/SetSpeedTestServer2?expireDate={expireDate}&port={port}&inBufferBytes={inBufferBytes}&outBufferBytes={outBufferBytes}&maxXferBytes={maxXferBytes}&timeout={timeout}&code1={code1}&code2={code2}";        
        public const int DEFAULT_NUM_THREADS = 1;
        public static string DEFAULT_LOCAL_IP_ADDRESS = IPAddress.Any.ToString();
        public const string DEFAULT_TITLE = "MmgSpeedTest";

        #endregion

        #region variables

        //LOCAL VARIABLES
        private int inBufferSize = DEFAULT_IN_BUFFER;
        private int outBufferSize = DEFAULT_OUT_BUFFER;

        private long serverRegistrationStart = -1;
        private long serverRegistrationNow = -1;
        private string version = "1.0.0.1";
        /// <summary>
        /// Application name for this windows service.
        /// </summary>
        private string appName = DEFAULT_TITLE;

        /// <summary>
        /// Application title for this windows service.
        /// </summary>
        private string title = DEFAULT_TITLE;

        /// <summary>
        /// The port this service will begin listening on.
        /// </summary>
        private int port = DEFAULT_PORT;

        /// <summary>
        /// Variable that tracks the current socket count.
        /// </summary>
        private int socketCount = 0;

        /// <summary>
        /// Constant that defines the number of threads to open up.
        /// </summary>
        private int numThreads = DEFAULT_NUM_THREADS;

        /// <summary>
        /// Constant that defines the connection timeout in ms.
        /// </summary>
        private int timeout = DEFAULT_TIMEOUT;

        /// <summary>
        /// Hashtable that tracks current connections.
        /// </summary>
        //private Hashtable connHt = new Hashtable();

        /// <summary>
        /// ArrayList that tracks the connection sockets.
        /// </summary>
        private ArrayList connSockets = null;

        /// <summary>
        /// Callback event used by the socket connection code.
        /// </summary>
        private ManualResetEvent allDone = new ManualResetEvent(false);

        /// <summary>
        /// Main thread pool for connection management.
        /// </summary>
        private Thread[] serverThread;

        /// <summary>
        /// Thread end event callback pool.
        /// </summary>
        //private AutoResetEvent[] threadEnd;

        /// <summary>
        /// Boolean that indicates the shutdown state of the service.
        /// </summary>
        private bool shuttingDown = false;

        /// <summary>
        /// The maximum allowed number of sockets.
        /// </summary>
        private int maxSockets = DEFAULT_MAX_SOCKETS;

        private int oldConnMin = DEFAULT_OLD_CONN_MIN;

        /// <summary>
        /// The root directory used for config files and log files.
        /// </summary>
        private string rootDir = System.Environment.GetFolderPath(System.Environment.SpecialFolder.ApplicationData);

        /// <summary>
        /// A variable that holds the application directory for this service.
        /// </summary>
        private string appDir = "";

        /// <summary>
        /// The expected file name for the config file.
        /// </summary>
        private string configFileName = "config.txt";

        /// <summary>
        /// The full path to the config file.
        /// </summary>
        private string configFilePath = "";

        /// <summary>
        /// The expected file name for the log file.
        /// </summary>
        private string debugFileName = "debug.txt";

        /// <summary>
        /// The full path to the log file.
        /// </summary>
        private string debugFilePath = "";

        /// <summary>
        /// Boolean that determines if logging is on or off.
        /// </summary>
        private bool debugOn = DEFAULT_DEBUG_ON;

        private bool publicOn = DEFAULT_PUBLIC_ON;

        /// <summary>
        /// Static local reference.
        /// </summary>
        public static SpeedTest cApp;

        /// <summary>
        /// Log writer.
        /// </summary>
        private StreamWriter debug = null;

        /// <summary>
        /// Check log size tick.
        /// </summary>
        private int wrCount = 0;

        private string speedTestRegUrl = DEFAULT_SPEED_TEST_REGISTRATION_URL;

        private int maxXferSize = DEFAULT_MAX_XFER;
        private string localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
        private string code1 = "middlemindgames.com-register-speed-test";
        private string code2 = "middlemindgames.com-speed-test-2";

        #endregion

        #region constructors

        /// <summary>
        /// Default constructure expects a port number and a title.
        /// The port number can be overwritten by the config file.
        /// </summary>
        /// <param name="Port"></param>
        /// <param name="Title"></param>
        public SpeedTest(int Port, string Title)
        {
            cApp = this;

            appDir = Path.Combine(rootDir, appName);
            configFilePath = Path.Combine(appDir, configFileName);
            debugFilePath = Path.Combine(appDir, debugFileName);

            Console.WriteLine("");
            Console.WriteLine("App Dir: " + appDir);
            Console.WriteLine("Config Path: " + configFilePath);
            Console.WriteLine("Debug Path: " + debugFilePath);

            LoadSettings();
            CreateDebugFile();
            BeginRegisterServer();
            
            serverThread = new Thread[numThreads];
            connSockets = new ArrayList(maxSockets);
        }

        #endregion

        private void BeginRegisterServer()
        {
            wr("BeginRegisterServer");
            ThreadStart ts = new ThreadStart(RegisterServerWebCall);
            Thread trd = new Thread(ts);
            trd.Start();
        }

        private void RegisterServerWebCall()
        {
            wr("RegisterServerWebCall");
            //HTTP GET call to the registration server
            if (speedTestRegUrl != null && speedTestRegUrl != "")
            {
                bool pass = false;
                long diff = 0;
                TimeSpan ts;

                if (serverRegistrationStart == -1)
                {
                    pass = true;
                }
                else
                {
                    serverRegistrationNow = System.DateTime.Now.Ticks;
                    diff = serverRegistrationNow - serverRegistrationStart;
                    ts = new TimeSpan(diff);
                    if (ts.Days >= 1)
                    {
                        pass = true;
                    }
                    else
                    {
                        pass = false;
                    }
                }

                if (pass == true)
                {
                    try
                    {
                        serverRegistrationStart = System.DateTime.Now.Ticks;
                        serverRegistrationNow = -1;

                        string url = (string)speedTestRegUrl.Clone();
                        //url = url.Replace("{expireDate}", Uri.EscapeDataString(System.DateTime.Now.AddDays(1.0).ToString("MM-dd-yyyy") + " 00:00:00:000"));

                        string expDateStr = Uri.EscapeDataString(System.DateTime.Now.AddDays(1.0).ToString("MM-dd-yyyy"));
                        wr("Found expiration date: " + expDateStr);

                        url = url.Replace("{expireDate}", expDateStr);
                        url = url.Replace("{port}", Uri.EscapeDataString((port + "")));
                        url = url.Replace("{inBufferBytes}", Uri.EscapeDataString((inBufferSize + "")));
                        url = url.Replace("{outBufferBytes}", Uri.EscapeDataString((outBufferSize + "")));
                        url = url.Replace("{maxXferBytes}", Uri.EscapeDataString((maxXferSize + "")));
                        url = url.Replace("{timeout}", Uri.EscapeDataString((timeout + "")));
                        url = url.Replace("{code1}", Uri.EscapeDataString(code1));
                        url = url.Replace("{code2}", Uri.EscapeDataString(code2));

                        WebClient wc = new WebClient();
                        wr("Registration Url: " + url);
                        string res = wc.DownloadString(url);
                        wr("Registration Response: " + res);
                    }
                    catch (Exception ex)
                    {
                        wr(ex.Message);
                        wr(ex.StackTrace);
                    }
                }
            }
        }


        #region create_log_file

        /// <summary>
        /// Opens up a new logging file.
        /// </summary>
        private void CreateDebugFile()
        {
            CreateDebugFile(true);
        }

        /// <summary>
        /// Opens up a new logging file with the append setting.
        /// </summary>
        /// <param name="append"></param>
        private void CreateDebugFile(bool append)
        {
            if (this.debugOn == true)
            {
                this.debug = new StreamWriter(this.debugFilePath, append);
                this.debug.AutoFlush = true;
                this.wrCount = 0;
            }
        }

        public void CloseDebugFile()
        {
            try
            {
                this.debug.Flush();
            }
            catch { }

            try
            {
                this.debug.Close();
            }
            catch { }
        }

        #endregion

        #region get_methods

        /// <summary>
        /// Returns the title of the service that has been set.
        /// </summary>
        /// <returns></returns>
        public string GetTitle()
        {
            return title;
        }

        /// <summary>
        /// Returns true if logging is turned on.
        /// </summary>
        /// <returns></returns>
        public bool GetDebugOn()
        {
            return debugOn;
        }

        /// <summary>
        /// Returns the full path to the logging file.
        /// </summary>
        /// <returns></returns>
        public string GetDebugFilePath()
        {
            return debugFilePath;
        }

        /// <summary>
        /// Returns the file name of the logging file.
        /// </summary>
        /// <returns></returns>
        public string GetDebugFileName()
        {
            return debugFileName;
        }

        /// <summary>
        /// Returns the application name of this service.
        /// </summary>
        /// <returns></returns>
        public string GetAppName()
        {
            return appName;
        }

        /// <summary>
        /// Returns the port number this service is using.
        /// </summary>
        /// <returns></returns>
        public int GetPort()
        {
            return port;
        }

        /// <summary>
        /// Returns the socket count of this service.
        /// </summary>
        /// <returns></returns>
        public int GetSocketCount()
        {
            return socketCount;
        }

        /// <summary>
        /// Returns the full path of the root directory the service is using.
        /// </summary>
        /// <returns></returns>
        public string GetRootDir()
        {
            return rootDir;
        }

        /// <summary>
        /// Returns the full path of the service config/log directory.
        /// </summary>
        /// <returns></returns>
        public string GetAppDir()
        {
            return appDir;
        }

        /// <summary>
        /// Returns the name of the config file this application is using.
        /// </summary>
        /// <returns></returns>
        public string GetConfigFileName()
        {
            return configFileName;
        }

        /// <summary>
        /// Returns the full path of the expected config file.
        /// </summary>
        /// <returns></returns>
        public string GetConfigFilePath()
        {
            return configFilePath;
        }

        /// <summary>
        /// Returns true if this service is shutting down.
        /// </summary>
        /// <returns></returns>
        public bool GetShuttingDown()
        {
            return shuttingDown;
        }

        #endregion

        /// <summary>
        /// Attempts to read the expected config file and load the settings.
        /// Running the service in cmd line mode shows the expected file paths and config file format.
        /// </summary>
        private void LoadSettings()
        {
            Console.WriteLine("");
            Console.WriteLine("Loading settings...");
            Console.WriteLine("Looking for data directory: " + this.appDir);
            if(Directory.Exists(this.appDir) == false)
            {
                Console.WriteLine("Creating the app directory, this is where the config file should go, config.txt.");
                Directory.CreateDirectory(this.appDir);
            }

            Console.WriteLine("");
            Console.WriteLine("Looking for config file: " + this.configFilePath);
            Console.WriteLine("Entries in the config file are as follows.");
            Console.WriteLine("port=[" + DEFAULT_PORT + ": The desired port number.]");
            Console.WriteLine("debug=[" + (DEFAULT_DEBUG_ON ? "1" : "0") + ": Set to 1/true for true, 0/false for false.]");
            Console.WriteLine("max_sockets=[" + DEFAULT_MAX_SOCKETS + ": An integer value representing the number of connections, use a small number if your server has few resources.]");
            Console.WriteLine("public=[" + (DEFAULT_PUBLIC_ON ? "1" : "0") + ": Set to 1/true for true, 0/false for false. Will register server once every 24 hours.]");
            Console.WriteLine("connection_timeout_ms=[" + DEFAULT_TIMEOUT + ": The number of miliseconds to wait for socket timeout.]");
            Console.WriteLine("old_socket_kill_min=[5: The number of minutes to wait before killing old sockets.]");
            Console.WriteLine("speed_test_registration_url:[" + DEFAULT_SPEED_TEST_REGISTRATION_URL + ": The speed test server registration URL.]");
            Console.WriteLine("in_buffer_size=[" + DEFAULT_IN_BUFFER + ": The number of bytes for the input connection buffer.]");
            Console.WriteLine("out_buffer_size=[" + DEFAULT_OUT_BUFFER + ": The number of bytes for the output connection buffer.]");
            Console.WriteLine("max_xfer_size=[" + DEFAULT_MAX_XFER + ": The max number of bytes for the entire tranfer.]");
            //Console.WriteLine("num_threads=[" + DEFAULT_NUM_THREADS + ": The number of threads the server should run with.]");
            Console.WriteLine("local_ip_address=[" + DEFAULT_LOCAL_IP_ADDRESS + ": The local IP address to bind to.]");

            Console.WriteLine("");
            if(File.Exists(this.configFilePath) == true)
            {
                #region config_exists

                Console.WriteLine("Found a config file.");
                Console.WriteLine("Loading config file values...");
                StreamReader sr = new StreamReader(this.configFilePath);
                string line = sr.ReadLine();
                while(line != null)
                {
                    bool success = false;
                    int v = 0;
                    IPAddress ip;
                    string tmp = "";
                    string[] pair = line.Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                    if(pair.Length == 2)
                    {
                        string key = pair[0];
                        string val = pair[1];
                        if(key.ToUpper() == "PORT")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if(success == true)
                            {
                                if(port >= MIN_PORT && port <= MAX_PORT)
                                {
                                    port = v;
                                    Console.WriteLine("Found port: " + this.port);
                                }
                                else
                                {
                                    port = DEFAULT_PORT;
                                    Console.WriteLine("Port outside bounds, using default timeout: " + this.port);
                                }
                            }
                            else
                            {
                                port = DEFAULT_PORT;
                                Console.WriteLine("Could not parse port, using default port: " + this.port);
                            }
                        }
                        else if(key.ToUpper() == "DEBUG")
                        {
                            tmp = val.ToUpper();
                            if(tmp == "1" || tmp == "TRUE")
                            {
                                this.debugOn = true;
                            }
                            else
                            {
                                this.debugOn = false;
                            }
                            Console.WriteLine("Found debug: " + this.debugOn);
                        }
                        else if (key.ToUpper() == "MAX_SOCKETS")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_SOCKETS && v <= MAX_SOCKETS)
                                {
                                    maxSockets = v;
                                    Console.WriteLine("Found max sockets: " + this.maxSockets);
                                }
                                else
                                {
                                    maxSockets = DEFAULT_MAX_SOCKETS;
                                    Console.WriteLine("Max Sockets outside bounds, using default max sockets: : " + this.maxSockets);
                                }
                            }
                            else
                            {
                                maxSockets = DEFAULT_MAX_SOCKETS;
                                Console.WriteLine("Could not parse max sockets, using default max sockets: " + this.maxSockets);
                            }
                        }
                        else if (key.ToUpper() == "NUM_THREADS")
                        {
                            /*
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_NUM_THREADS && v <= MAX_NUM_THREADS)
                                {
                                    numThreads = v;
                                    Console.WriteLine("Found num threads: " + this.numThreads);
                                }
                                else
                                {
                                    numThreads = DEFAULT_NUM_THREADS;
                                    Console.WriteLine("Num threads outside bounds, using default num threads: " + this.numThreads);
                                }
                            }
                            else
                            {
                            */ 
                                numThreads = DEFAULT_NUM_THREADS;
                                Console.WriteLine("Could not parse num threads, using default num threads: " + this.numThreads);
                            //}
                        }
                        else if (key.ToUpper() == "PUBLIC")
                        {
                            tmp = val.ToUpper();
                            if (tmp == "1" || tmp == "TRUE")
                            {
                                this.publicOn = true;
                            }
                            else
                            {
                                this.publicOn = false;
                            }
                            Console.WriteLine("Found public: " + this.publicOn);
                        }
                        else if (key.ToUpper() == "CONNECTION_TIMEOUT_MS")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_TIMEOUT && v <= MAX_TIMEOUT)
                                {
                                    timeout = v;
                                    Console.WriteLine("Found timeout: " + this.timeout);
                                }
                                else
                                {
                                    timeout = DEFAULT_TIMEOUT;
                                    Console.WriteLine("Timeout outside bounds, using default timeout: " + this.timeout);
                                }
                            }
                            else
                            {
                                timeout = DEFAULT_TIMEOUT;
                                Console.WriteLine("Could not parse timeout, using default timeout: " + this.timeout);
                            }
                        }
                        else if (key.ToUpper() == "OLD_SOCKET_KILL_MIN")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_KILL_MIN && v <= MAX_KILL_MIN)
                                {
                                    oldConnMin = v;
                                    Console.WriteLine("Found old socket kill min: " + this.oldConnMin);
                                }
                                else
                                {
                                    oldConnMin = DEFAULT_OLD_CONN_MIN;
                                    Console.WriteLine("Old conn min outside bounds, using default old conn min: " + this.oldConnMin);
                                }
                            }
                            else
                            {
                                oldConnMin = DEFAULT_OLD_CONN_MIN;
                                Console.WriteLine("Could not parse old socket kill min, using default old socket kill min: " + this.oldConnMin);
                            }
                        }
                        else if (key.ToUpper() == "OUT_BUFFER_SIZE")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_OUT_BUFFER_LENGTH && v <= MAX_OUT_BUFFER_LENGTH)
                                {
                                    outBufferSize = v;
                                    Console.WriteLine("Found out buffer size: " + this.outBufferSize);
                                }
                                else
                                {
                                    outBufferSize = DEFAULT_OUT_BUFFER;
                                    Console.WriteLine("Out buffer size outside bounds, using default out buffer size: " + this.outBufferSize);
                                }
                            }
                            else
                            {
                                outBufferSize = DEFAULT_OUT_BUFFER;
                                Console.WriteLine("Could not parse out buffer size, using default out buffer size: " + this.outBufferSize);
                            }
                        }
                        else if (key.ToUpper() == "IN_BUFFER_SIZE")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_IN_BUFFER_LENGTH && v <= MAX_IN_BUFFER_LENGTH)
                                {
                                    inBufferSize = v;
                                    Console.WriteLine("Found in buffer size: " + this.inBufferSize);
                                }
                                else
                                {
                                    inBufferSize = DEFAULT_IN_BUFFER;
                                    Console.WriteLine("In buffer size outside bounds, using default in buffer size: " + this.inBufferSize);
                                }
                            }
                            else
                            {
                                inBufferSize = DEFAULT_IN_BUFFER;
                                Console.WriteLine("Could not parse in buffer size, using default in buffer size: " + this.inBufferSize);
                            }
                        }
                        else if (key.ToUpper() == "MAX_XFER_SIZE")
                        {
                            success = false;
                            success = Int32.TryParse(val, out v);
                            if (success == true)
                            {
                                if (v >= MIN_XFER_LIMIT && v <= MAX_XFER_LIMIT)
                                {
                                    maxXferSize = v;
                                    Console.WriteLine("Found max xfer size: " + this.maxXferSize);
                                }
                                else
                                {
                                    maxXferSize = DEFAULT_MAX_XFER;
                                    Console.WriteLine("Max xfer size outside bounds, using default max xfer size: " + this.maxXferSize);
                                }
                            }
                            else
                            {
                                maxXferSize = DEFAULT_MAX_XFER;
                                Console.WriteLine("Could not parse max xfer size, using default max xfer size: " + this.maxXferSize);
                            }
                        }
                        else if(key.ToUpper() == "SPEED_TEST_REGISTRATION_URL")
                        {
                            if (val != null && val != "")
                            {
                                speedTestRegUrl = val;
                                Console.WriteLine("Found speed test registration url: " + this.maxXferSize);
                            }
                            else
                            {
                                speedTestRegUrl = DEFAULT_SPEED_TEST_REGISTRATION_URL;
                                Console.WriteLine("Could not parse speed test registration url, using default speed test registration url: " + this.speedTestRegUrl);
                            }
                        }
                        else if (key.ToUpper() == "LOCAL_IP_ADDRESS")
                        {
                            success = false;
                            success = IPAddress.TryParse(val, out ip);
                            if (success == true)
                            {
                                localIpAddress = val;
                                Console.WriteLine("Found local ip address: " + this.localIpAddress);
                            }
                            else
                            {
                                localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
                                Console.WriteLine("Could not parse local ip address, using default local ip address: " + this.localIpAddress);
                            }
                        }
                    }
                    line = sr.ReadLine();
                }

                #endregion
            }
            else
            {
                Console.WriteLine("Could not find a config file.");
                Console.WriteLine("Loading default values...");
                port = DEFAULT_PORT;
                debugOn = DEFAULT_DEBUG_ON;
                maxSockets = DEFAULT_MAX_SOCKETS;
                publicOn = DEFAULT_PUBLIC_ON;
                timeout = DEFAULT_TIMEOUT;
                oldConnMin = DEFAULT_OLD_CONN_MIN;
                speedTestRegUrl = DEFAULT_SPEED_TEST_REGISTRATION_URL;
                outBufferSize = DEFAULT_OUT_BUFFER;
                inBufferSize = DEFAULT_IN_BUFFER;
                maxXferSize = DEFAULT_MAX_XFER;
                numThreads = DEFAULT_NUM_THREADS;
                localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
            }

            PrintConfig();
        }

        public void PrintConfig()
        {
            Console.WriteLine("==========CURRENT CONFIG===========");
            Console.WriteLine("Version: " + version);
            Console.WriteLine("Found port: " + this.port);
            Console.WriteLine("Found debug: " + this.debugOn);
            Console.WriteLine("Found max_sockets: " + this.maxSockets);
            Console.WriteLine("Found public: " + this.publicOn);
            Console.WriteLine("Found connection_timeout_ms: " + this.timeout + " ms");
            Console.WriteLine("Found old_socket_kill_min: " + this.oldConnMin + " min");
            Console.WriteLine("Found speed_test_registration_url: " + this.speedTestRegUrl);
            Console.WriteLine("Found in_buffer_size: " + this.inBufferSize + " bytes");
            Console.WriteLine("Found out_buffer_size: " + this.outBufferSize + " bytes");
            Console.WriteLine("Found max_xfer_size: " + this.maxXferSize + " bytes");
            Console.WriteLine("Found num_threads: " + this.numThreads);
            Console.WriteLine("Found local_ip_address: " + this.localIpAddress);

            if (System.Net.NetworkInformation.NetworkInterface.GetIsNetworkAvailable() == true)
            {
                Console.WriteLine("Found local ip pool: ");
                this.GetLocalIPAddress();
            }
            else
            {
                Console.WriteLine("Network connection not found.");
            }
        }

        public void GetLocalIPAddress()
        {
            var host = Dns.GetHostEntry(Dns.GetHostName());
            int count = 1;
            foreach (var ip in host.AddressList)
            {
                if (ip.AddressFamily == AddressFamily.InterNetwork)
                {
                    wr(count + ": " + ip.ToString());
                    count++;
                }
            }
        }

        public void PrintState()
        {
            Console.WriteLine("==========CURRENT STATE===========");
            Console.WriteLine("Socket Count: " + GetSocketCount());
            Console.WriteLine("Max Sockets: " + this.maxSockets);
        }

        /// <summary>
        /// Closes all current connections.
        /// </summary>
        private void CloseAllConnections()
        {
            try
            {
                //Clean up all connected sockets.
                for (int i = 0; i < socketCount; i++)
                {
                    StateObject st = (StateObject)this.connSockets[i];
                    st.workSocket.Close();
                    connSockets[i] = null;
                }
            }
            catch (Exception e)
            {
                wr("Error in close all connections." + e.Message);
            }
        }

        /// <summary>
        /// Closes all old connections.
        /// </summary>
        private void CloseAllOldConnections()
        {
            try
            {
                bool kill = false;
                for (int i = 0; i < socketCount; i++)
                {
                    StateObject st = (StateObject)connSockets[i];

                    try
                    {
                        kill = false;
                        int n = st.workSocket.Available;
                    }
                    catch
                    {
                        kill = true;
                    }


                    if (st != null && st.workSocket != null && (System.DateTime.Now.Ticks - st.connTime) > (TimeSpan.TicksPerMinute * this.oldConnMin) || st.workSocket.Connected == false || kill == true)
                    {
                        try
                        {
                            wr("##### Removing socket at index: " + i);
                            RemoveSocket(st);
                        }
                        catch (Exception e)
                        {
                            wr("Error closing connection 1.");
                            wr(e.Message);
                            wr(e.StackTrace);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                wr("Error in close all old connections." + e.Message);
            }
        }

        /// <summary>
        /// Stop the service and shut it down.
        /// </summary>
        public void Stop()
        {
            shuttingDown = true;
        }

        /// <summary>
        /// Start the service listening.
        /// </summary>
        public void Start()
        {
            wr("Starting: " + title);

            /*
            //Clear all thread events.
            for (int i = 0; i < numThreads; i++)
            {
                threadEnd[i] = new AutoResetEvent(false);
            }
            */
            ThreadStart starter = new ThreadStart(StartListening);
            serverThread[0] = new Thread(starter);
            serverThread[0].IsBackground = true;
            serverThread[0].Start();
        }

        /// <summary>
        /// Write to the console and the log file if debugging is turned on.
        /// </summary>
        /// <param name="s"></param>
        private void wr(string s)
        {
            if (this.debugOn == true)
            {
                Console.WriteLine(System.DateTime.Now.Ticks + ": " + s);
                if (this.debug != null)
                {
                    if(s == null)
                    {
                        s = "";
                    }
                    
                    this.debug.WriteLine(s);
                    this.debug.Flush();
                    wrCount++;
                }

                if (wrCount > 0 && wrCount % CHECK_LOG_SIZE_TICKS == 0)
                {
                    FileInfo fInf = new FileInfo(this.debugFilePath);
                    if (fInf.Length >= CHECK_LOG_SIZE)
                    {
                        this.debug.Flush();
                        this.debug.Close();
                        CreateDebugFile(false);
                    }
                }
            }
        }

        /// <summary>
        /// Begin listening for connection requests.
        /// </summary>
        private void StartListening()
        {
            wr("StartListening");
            //Establish the local end point for the socket.
            IPEndPoint endPoint = new IPEndPoint(IPAddress.Any, port);

            //Create TCP/IP socket.
            Socket listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            listener.SendTimeout = timeout;
            listener.ReceiveTimeout = timeout;

            //Bind the socket to the local end point.
            try
            {
                listener.Bind(endPoint);
                listener.Listen((Int32)500);
                wr("Listening on port: " + port + " Local IP: " + listener.LocalEndPoint);

                while (!shuttingDown)
                {
                    //Sets the event to a non signaled state
                    allDone.Reset();

                    //Start asynchronous socket to listen for connections.
                    listener.BeginAccept(new AsyncCallback(AcceptCallback), listener);

                    //Wait until a connection is made before continuing.
                    wr("waiting...");
                    allDone.WaitOne();
                }
            }
            catch (Exception e)
            {
                wr("Error caught in StartListening.");
                wr(e.Message);
                wr(e.StackTrace);
                //threadEnd[0].Set();
            }
        }

        /// <summary>
        /// Callback method used when a new connection is received.
        /// </summary>
        /// <param name="ar"></param>
        private void AcceptCallback(IAsyncResult ar)
        {
            //Signal the main thread to continue.
            allDone.Set();

            //Get the socket that handles the client request.
            Socket listener = (Socket)ar.AsyncState;
            Socket handler = listener.EndAccept(ar);

            wr("Socket receive buffer size 1: " + handler.ReceiveBufferSize);

            //Create the state object.
            StateObject state = new StateObject(inBufferSize, outBufferSize);
            state.connTime = System.DateTime.Now.Ticks;
            state.workSocket = handler;
            handler.ReceiveBufferSize = inBufferSize;
            handler.SendBufferSize = outBufferSize;
            handler.SendTimeout = timeout;
            handler.ReceiveTimeout = timeout;
            handler.NoDelay = true;
            state.timeStamp = DateTime.Now;

            wr("Socket receive buffer size 2: " + handler.ReceiveBufferSize);
            wr("Found connection request on " + System.DateTime.Now.Ticks + " from " + state.workSocket.RemoteEndPoint);

            if (publicOn == true)
            {
                BeginRegisterServer();
            }

            try
            {
                if (socketCount < maxSockets)
                {
                    Interlocked.Increment(ref socketCount);
                    
                    Monitor.Enter(connSockets);
                    connSockets.Add(state);
                    Monitor.Exit(connSockets);

                    handler.BeginReceive(state.inBuffer, 0, inBufferSize, SocketFlags.None, new AsyncCallback(NetworkReadCallBack), state);
                }
                else
                {
                    wr("Socket count has reached the max. Cannot process request.");
                }
            }
            catch (Exception e)
            {
                wr("Error caught in AcceptCallback, trying to close this connection.");
                wr(e.Message);
                wr(e.StackTrace);
                try
                {
                    RemoveSocket(state);
                }
                catch { }
            }
        }

        /// <summary>
        /// Legacy, doesn't alter the b2 parameter but is available for data compression/encoding.
        /// </summary>
        /// <param name="b2"></param>
        /// <returns></returns>
        private byte[] EncodeData(byte[] b2)
        {
            //wr("Src Length: " + b2.Length);
            return b2;
        }

        /// <summary>
        /// Passes back b2 unless a shorten request is made then b2 is capped to the passed in length.
        /// </summary>
        /// <param name="b2"></param>
        /// <param name="length"></param>
        /// <param name="shorten"></param>
        /// <returns></returns>
        private byte[] DecodeData(byte[] b2, int length, bool shorten)
        {
            // decode input arguments
            byte[] b3 = null;

            if (shorten)
            {
                byte[] b1 = new byte[length];
                System.Array.Copy(b2, 0, b1, 0, length);
                b3 = b1;
            }
            else
            {
                b3 = b2;
            }

            return b3;
        }

        /// <summary>
        /// Legacy, gets a string from a byte array, no encoding is provided here.
        /// THis function is not currently used but exists as a compliment to BytesFromString.
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        private string StringFromBytes(byte[] b)
        {
            char[] ca = new char[b.Length];
            System.Array.Copy(b, 0, ca, 0, ca.Length);
            return new String(ca);
        }

        /// <summary>
        /// Returns the bytes from the given string using UTF8 encoding.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        private byte[] BytesFromString(string s)
        {
            return System.Text.Encoding.UTF8.GetBytes(s);
        }

        private void CloseSocket(StateObject state, Socket handler)
        {
            wr("Disconnecting");
            try
            {
                handler.Disconnect(true);
            }
            catch (Exception ex3)
            {
                wr("Error caught in NetworkReadCallBack server disconnect call - Handler Disconnect");
                wr(ex3.Message);
                wr(ex3.StackTrace);
            }

            try
            {
                RemoveSocket(state);
            }
            catch (Exception ex3)
            {
                wr("Error caught in NetworkReadCallBack server disconnect call - RemoveSocket");
                wr(ex3.Message);
                wr(ex3.StackTrace);
            }

            try
            {
                CloseAllOldConnections();
            }
            catch (Exception ex3)
            {
                wr("Error caught in NetworkReadCallBack server disconnect call - CloseAllOldConnections.");
                wr(ex3.Message);
                wr(ex3.StackTrace);
            }

            this.PrintState();
        }

        /// <summary>
        /// Handles network read callback requests.
        /// </summary>
        /// <param name="ar"></param>
        private void NetworkReadCallBack(IAsyncResult ar)
        {
            //Retrieve the state object and the handler socket.
            wr("");
            wr("");
            wr("NeworkReadCallBack called ");
            StateObject state = (StateObject)ar.AsyncState;
            Socket handler = null;

            lock (state)
            {
                try
                {
                    handler = state.workSocket;
                    if (handler == null)
                    {
                        return;
                    }

                    wr("Read call back request from " + state.workSocket.RemoteEndPoint);

                    //Read data from the client socket.
                    int bytesRead = handler.EndReceive(ar);
                    if (bytesRead > 0)
                    {
                        wr("Bytes Read: " + bytesRead + " Bytes received: " + state.received + " Bytes Expecting: " + state.expecting + " " + System.DateTime.Now.Ticks);
                        string code = System.Text.Encoding.UTF8.GetString(state.inBuffer, 0, 3);
                        byte[] b2 = null;
                        byte[] b3 = null;
                        int fullLen = 0;
                        int xferLen = 0;
                        int tlen = 0;
                        int count = 0;
                        wr("Found Code: " + code);

                        switch (code)
                        {
                            case "0??":
                                //UPLOAD FILE
                                wr("UPLOAD FILE: 0??");
                                wr("Bytesread: " + bytesRead);

                                b2 = new byte[4];
                                System.Array.Copy(state.inBuffer, 3, b2, 0, 4);
                                state.expecting = net.jbcg.Encryption.Converter.toInt(b2);

                                b3 = new byte[4];
                                System.Array.Copy(b2, b3, 4);

                                if (state.expecting <= 0 || state.expecting > (MiB_INT * 100))
                                {
                                    throw new Exception("Illegal array size: " + state.expecting);
                                }
                                state.received = 0;

                                if (bytesRead >= 7)
                                {
                                    state.received = (bytesRead - 7);
                                }

                                wr("Expecting: " + state.expecting + ", Received: " + state.received);
                                int tries = 0;
                                int maxTries = (2048 * 2);
                                while (state.received < state.expecting)
                                {
                                    int dIn = state.workSocket.Receive(state.inBuffer);
                                    if (dIn == -1)
                                    {
                                        wr("Found -1 in socket read...exiting....");
                                        break;
                                    }

                                    state.received += dIn;
                                    //wr("Found " + dIn + " bytes, total " + state.received + " bytes, expecting " + state.expecting + " bytes");

                                    if (tries > maxTries)
                                    {
                                        wr("Error, maximum tries reached.");
                                        break;
                                    }
                                    tries++;
                                }

                                if (state.received < state.expecting)
                                {
                                    wr("Found " + state.received + ", expecting " + state.expecting + ", error.");
                                    //return;
                                }

                                wr("DONE: total " + state.received + " bytes, expecting " + state.expecting + " bytes");
                                state.expecting = 0;
                                state.received = 0;
                                CloseSocket(state, handler);
                                break;

                            #region DOWNLOAD_1KB_TO_100KB
                            case "1??":
                                //DOWNLOAD 1K FILE
                                byte[] bb2 = null;
                                byte[] bb3 = null;

                                bb2 = new byte[KiB_INT];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;

                            case "2??":
                                //DOWNLOAD 100KB FILE
                                bb2 = new byte[KiB_INT * 100];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;
                            #endregion

                            case "3??":
                                //SERVER DISCONNECT
                                CloseSocket(state, handler);
                                return;

                            #region DOWNLOAD_50KB_TO_1MB
                            case "4??":
                                //DOWNLOAD 50KB FILE
                                bb2 = new byte[KiB_INT * 50];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;

                            case "5??":
                                //DOWNLOAD 10K FILE
                                bb2 = new byte[KiB_INT * 10];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;

                            case "6??":
                                //DOWNLOAD 1MB FILE
                                bb2 = new byte[MiB_INT];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;
                            #endregion

                            #region DOWNLOAD_5MB_TO_50MB
                            case "7??":
                                //DOWNLOAD 5MB FILE
                                bb2 = new byte[MiB_INT * 5];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(bb2.Length);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);
                                state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;

                            case "8??":
                                //DOWNLOAD 10MB FILE
                                fullLen = MiB_INT * 10;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if(count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "9??":
                                //DOWNLOAD 15MB FILE
                                fullLen = MiB_INT * 15;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "10?":
                                //DOWNLOAD 20MB FILE
                                fullLen = MiB_INT * 20;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "11?":
                                //DOWNLOAD 25MB FILE
                                fullLen = MiB_INT * 25;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "12?":
                                //DOWNLOAD 30MB FILE
                                fullLen = MiB_INT * 30;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "13?":
                                //DOWNLOAD 35MB FILE
                                fullLen = MiB_INT * 35;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "14?":
                                //DOWNLOAD 40MB FILE
                                fullLen = MiB_INT * 40;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "15?":
                                //DOWNLOAD 45MB FILE
                                fullLen = MiB_INT * 45;
                                xferLen = outBufferSize;
                                tlen = (fullLen / xferLen) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;

                            case "16?":
                                //DOWNLOAD 50MB FILE
                                fullLen = MiB_INT * 50;
                                xferLen = outBufferSize;
                                tlen = (fullLen / outBufferSize) + 1;

                                bb2 = new byte[xferLen];
                                bb2 = EncodeData(bb2);

                                bb3 = net.jbcg.Encryption.Converter.toByte(fullLen);
                                bb3 = EncodeData(bb3);

                                state.workSocket.Send(bb3, 0, bb3.Length, SocketFlags.None);

                                count = 0;
                                for (int i = 0; i < tlen; i++)
                                {
                                    if (count + outBufferSize > fullLen)
                                    {
                                        xferLen = (fullLen - count);
                                    }
                                    else
                                    {
                                        xferLen = outBufferSize;
                                    }

                                    bb2 = new byte[xferLen];
                                    bb2 = EncodeData(bb2);

                                    state.workSocket.Send(bb2, 0, bb2.Length, SocketFlags.None);

                                    count += bb2.Length;
                                }

                                CloseSocket(state, handler);
                                break;
                            #endregion

                            default:
                                //DEFAULT
                                wr("Empty code! " + (int)state.inBuffer[0] + " - " + (int)state.inBuffer[1] + " - " + (int)state.inBuffer[2]);
                                wr("Bytesread: " + bytesRead);
                                b2 = BytesFromString(code);

                                state.workSocket.Send(b2, 0, b2.Length, SocketFlags.None);
                                CloseSocket(state, handler);
                                break;
                        }
                    }
                }
                catch (System.Net.Sockets.SocketException es)
                {
                    wr("Socket exception caught in NetworkReadCallBack, removing socket.");
                    wr(es.Message);
                    wr(es.StackTrace);

                    wr("Remove socket 1.");
                    try
                    {
                        RemoveSocket(state);
                    }
                    catch(Exception ex3)
                    {
                        wr("Inner exception caught removing socket.");
                        wr(ex3.Message);
                        wr(ex3.StackTrace);
                    }
                }
                catch (Exception e)
                {
                    wr("Exception caught in NetworkReadCallBack, removing socket.");
                    wr(e.Message);
                    wr(e.StackTrace);

                    wr("Remove socket 2.");
                    try
                    {
                        RemoveSocket(state);
                    }
                    catch (Exception ex3)
                    {
                        wr("Inner exception caught removing socket.");
                        wr(ex3.Message);
                        wr(ex3.StackTrace);
                    }
                }
            }
        }

        /// <summary>
        /// Network write callback.
        /// </summary>
        /// <param name="ar"></param>
        private void NetworkWriteCallBack(IAsyncResult ar)
        {
            //Retrieve the state object and the handler socket.
            /*
            wr("");
            wr("");
            wr("NetworkWriteCallBack");
            StateObject state = (StateObject)ar.AsyncState;
            Socket handler = null;

            lock (state)
            {
                try
                {
                    handler = state.workSocket;

                    //Read data from the client socket.
                    int bytesSent = handler.EndSend(ar);
                    wr("Found bytes sent: " + bytesSent);

                    if (bytesSent > 0)
                    {
                        if (handler != null && handler.Connected)
                        {
                            wr("Begins receive...");
                            handler.BeginReceive(state.inBuffer, 0, inBufferSize, SocketFlags.None, new AsyncCallback(NetworkReadCallBack), state);
                        }
                    }

                    wr("Write call back request from " + state.workSocket.RemoteEndPoint);
                    wr("Sent " + bytesSent + " number of bytes.");
                }
                catch (System.Net.Sockets.SocketException es)
                {
                    wr("Socket exception caught in NetworkWriteCallBack, removing socket.");
                    wr(es.Message);
                    wr(es.StackTrace);

                    wr("Remove socket 1.");
                    try
                    {
                        RemoveSocket(state);
                    }
                    catch (Exception ex3)
                    {
                        wr("Inner exception caught removing socket.");
                        wr(ex3.Message);
                        wr(ex3.StackTrace);
                    }
                }
                catch (Exception e)
                {
                    wr("Exception caught in NetworkWriteCallBack, removing socket.");
                    wr(e.Message);
                    wr(e.StackTrace);

                    wr("Remove socket 2.");
                    try
                    {
                        RemoveSocket(state);
                    }
                    catch (Exception ex3)
                    {
                        wr("Inner exception caught removing socket.");
                        wr(ex3.Message);
                        wr(ex3.StackTrace);
                    }
                }
            }
            */
        }

        /// <summary>
        /// Removes a socket from the connected socket pool.
        /// </summary>
        /// <param name="state"></param>
        private void RemoveSocket(StateObject state)
        {
            try
            {
                wr("Removing socket");

                Socket sock = state.workSocket;
                Monitor.Enter(this.connSockets);
                if (this.connSockets.Contains(state))
                {
                    connSockets.Remove(state);
                    Interlocked.Decrement(ref this.socketCount);
                }
                Monitor.Exit(this.connSockets);

                if (sock != null)
                {
                    try
                    {
                        sock.Shutdown(SocketShutdown.Both);
                    }
                    catch { }

                    try
                    {
                        sock.Shutdown(SocketShutdown.Receive);
                    }
                    catch { }

                    try
                    {
                        sock.Shutdown(SocketShutdown.Send);
                    }
                    catch { }

                    try
                    {
                        sock.Close();
                    }
                    catch { }

                    sock = null;
                    state.workSocket = null;
                    state = null;
                }
            }
            catch (Exception e)
            {
                wr("Exception caught in the RemoveSocket method.");
                wr(e.Message);
                wr(e.StackTrace);
            }
        }
    }
}